import { keys, isUndefined, last, toString } from 'micro-dash';

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
/**
 * Extend this for classes whose objects are directly callable. Sadly, you'll need to define an extra interface and repeat the typing information, for now. (Maybe fixable with TypeScript 3 tuples?)
 *
 * ```ts
 * interface Multiplier {
 *   (value: number): number;
 * }
 *
 * class Multiplier extends CallableObject {
 *   constructor(public factor: number) {
 *     super((value: number) => value * this.factor);
 *   }
 * }
 *
 * const doubler = new Multiplier(2);
 * doubler(2); // 4
 * doubler.factor = 3;
 * doubler(2); // 6
 * ```
 * @abstract
 */
class CallableObject {
    /**
     * @param {?} f
     */
    constructor(f) {
        return Object.setPrototypeOf(f, new.target.prototype);
    }
}

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
/** @type {?} */
let seq = 0;
/**
 * Creates a factory function for building objects of a given type. Commonly used to create test objects for use in specs.
 *
 * ```ts
 * interface Message {
 *   id: number;
 *   text: string;
 * }
 *
 * interface Options {
 *   cypherDistance: number;
 * }
 *
 * declare function shiftCharacters(text: string, distance: number): string;
 *
 * const buildMessage = createBuilder<Message, Options>(
 *   (seq) => ({ id: seq, text: `message ${seq}` }),
 *   (message, _seq, options) => {
 *     message.text = shiftCharacters(message.text, options.cypherDistance || 0);
 *   },
 * );
 *
 * buildMessage(); // { id: 1, text: "message 1" }
 * buildMessage({ text: "Hello world!" }); // { id: 2, text: "Hello world!" }
 * buildMessage({ text: "abc" }, { cypherDistance: 3 }); // { id: 3, text: "def" }
 * ```
 * @template T, OptionsType
 * @param {?} buildDefaults
 * @param {?=} afterBuild
 * @return {?}
 */
function createBuilder(buildDefaults, afterBuild) {
    return (/**
     * @param {?=} attributes
     * @param {?=} options
     * @return {?}
     */
    (attributes, options = {}) => {
        ++seq;
        /** @type {?} */
        const obj = Object.assign(buildDefaults(seq, options), attributes);
        if (afterBuild) {
            afterBuild(obj, seq, options);
        }
        return obj;
    });
}

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
/**
 * An object you can use to create a promise and resolve/reject it later.
 *
 * ```ts
 * const deferred = new Deferred<string>();
 *
 * // later
 * deferred.resolve('a nice value'); // causes `deferred.promise` to resolve
 *
 * // or
 * deferred.reject('a bad value'); // causes `deferred.promise` to reject
 * ```
 * @template T
 */
class Deferred {
    constructor() {
        this.promise = new Promise((/**
         * @param {?} resolve
         * @param {?} reject
         * @return {?}
         */
        (resolve, reject) => {
            this.resolve = resolve;
            this.reject = reject;
        }));
    }
}

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
/**
 * Performs a deep comparison between two values to determine if they are equivalent, up to the given depth. Once that depth is reached, remaining values are compared using `Object.is()`.
 *
 * ```ts
 * let object1 = "a";
 * let object2 = "a";
 * isEqualAtDepth(0, object1, object2); // true
 * isEqualAtDepth(1, object1, object2); // true
 *
 * let object1 = { a: 1 };
 * let object2 = { a: 1 };
 * isEqualAtDepth(0, object1, object2); // false
 * isEqualAtDepth(1, object1, object2); // true
 *
 * object1 = { a: [1, 2, 3], d: { e: 1 } };
 * object2 = { a: [1, 2, 3], d: { e: 1 } };
 * isEqualAtDepth(0, object1, object2); // false
 * isEqualAtDepth(1, object1, object2); // false
 * isEqualAtDepth(2, object1, object2); // true
 * isEqualAtDepth(3, object1, object2); // true
 * ```
 * @param {?} depth
 * @param {?} value
 * @param {?} other
 * @return {?}
 */
function isEqualAtDepth(depth, value, other) {
    if (Object.is(value, other)) {
        // covers e.g. NaN === NaN
        return true;
    }
    if (depth === 0 || !(value instanceof Object && other instanceof Object)) {
        return false;
    }
    for (const key of keys(value)) {
        if (!other.hasOwnProperty(key)) {
            return false;
        }
    }
    for (const key of keys(other)) {
        if (!isEqualAtDepth(depth - 1, value[key], other[key])) {
            return false;
        }
    }
    return true;
}

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
/**
 * Checks if `value` is anything other than `undefined`.
 *
 * ```ts
 * isDefined({ a: 1 }); // true
 * isDefined(false); // true
 * isDefined(null); // true
 * isDefined(undefined); // false
 * ```
 * @param {?} value
 * @return {?}
 */
function isDefined(value) {
    return !isUndefined(value);
}

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
/**
 * Rounds `value` to the nearest multiple of `multiple`.
 *
 * ```ts
 * roundToMultiple(5, 13); // 15
 * roundToMultiple(2, 4.8); // 4
 * roundToMultiple(3, -4); // -3
 * roundToMultiple(0.5, 1.6); // 1.5
 * ```
 * @param {?} multiple
 * @param {?} value
 * @return {?}
 */
function roundToMultipleOf(multiple, value) {
    return Math.round(value / multiple) * multiple;
}

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
/**
 * Returns a promise that resolves after `ms` milliseconds.
 *
 * ```ts
 * // do something
 * await sleep(1000); // wait a second
 * // do something else
 * ```
 * @param {?} ms
 * @return {?}
 */
function sleep(ms) {
    return new Promise((/**
     * @param {?} resolve
     * @return {?}
     */
    (resolve) => {
        setTimeout(resolve, ms);
    }));
}

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
/** @type {?} */
const nanoConversions = {};
/** @enum {string} */
const TimeUnit = {
    Nanoseconds: "ns",
    Microseconds: "μs",
    Milliseconds: "ms",
    Seconds: "s",
    Minutes: "m",
    Hours: "h",
    Days: "d",
    Weeks: "w",
    Years: "y",
    Decades: "dec",
    Centuries: "cent",
    Millennia: "mil",
};
/**
 * Converts time between two units. Units can be any value described in the docs for [[TimeUnit]].
 *
 * ```ts
 * convertTime(1, 's', 'ms'); // 1000
 * convertTime(20, TimeUnit.Decades, TimeUnit.Centuries); // 2
 * ```
 * @param {?} value
 * @param {?} unit
 * @param {?} targetUnit
 * @return {?}
 */
function convertTime(value, unit, targetUnit) {
    return (value * nanoConversions[unit]) / nanoConversions[targetUnit];
}
/**
 * Constructs a string representation of an elapsed amount of time. The least significant unit will be rounded to the nearest whole number.
 *
 * ```ts
 * elapsedToString(2001, ["s", "ms"]); // "2 s 1 ms"
 * elapsedToString(15, ["wks", "d"], ( elapsedUnit: TimeUnit.Days }); // "2 wks 1 d"
 * elapsedToString(1, [TimeUnit.Microseconds]); // "1000 μs"
 * ```
 *
 * @param {?} elapsed
 * @param {?} units
 * @param {?=} __2
 * @return {?}
 */
function elapsedToString(elapsed, units, { elapsedUnit = TimeUnit.Milliseconds, showLeadingZeros = true } = {}) {
    elapsed = roundToMultipleOf(convertTime(1, last(units), elapsedUnit), elapsed);
    /** @type {?} */
    let showZeros = showLeadingZeros;
    /** @type {?} */
    const tokens = [];
    for (let i = 0; i < units.length; ++i) {
        /** @type {?} */
        const unit = units[i];
        /** @type {?} */
        const conversion = convertTime(1, elapsedUnit, unit);
        /** @type {?} */
        const value = Math.floor(elapsed * conversion);
        if (value > 0 || showZeros || i === units.length - 1) {
            tokens.push(value);
            tokens.push(unit);
            showZeros = true;
        }
        elapsed -= value / conversion;
    }
    return tokens.join(" ");
}
addConversion(TimeUnit.Nanoseconds, "Nanosecond", 1, { aliases: ["nanos"] });
addConversion(TimeUnit.Microseconds, "Microsecond", 1000, {
    aliases: ["micros"],
});
addConversion(TimeUnit.Milliseconds, "Millisecond", 1000 * 1000, {
    aliases: ["millis"],
});
addConversion(TimeUnit.Seconds, "Second", 1000 * 1000 * 1000, {
    aliases: ["S", "sec", "secs"],
});
addConversion(TimeUnit.Minutes, "Minute", 60 * 1000 * 1000 * 1000, {
    aliases: ["M", "min", "mins"],
});
addConversion(TimeUnit.Hours, "Hour", 60 * 60 * 1000 * 1000 * 1000, {
    aliases: ["H", "hr", "hrs"],
});
addConversion(TimeUnit.Days, "Day", 24 * 60 * 60 * 1000 * 1000 * 1000, {
    aliases: ["D"],
});
addConversion(TimeUnit.Weeks, "Week", 7 * 24 * 60 * 60 * 1000 * 1000 * 1000, {
    aliases: ["W", "wk", "wks"],
});
addConversion(TimeUnit.Years, "Year", 365 * 24 * 60 * 60 * 1000 * 1000 * 1000, {
    aliases: ["Y", "yr", "yrs"],
});
addConversion(TimeUnit.Decades, "Decade", 10 * 365 * 24 * 60 * 60 * 1000 * 1000 * 1000);
addConversion(TimeUnit.Centuries, "Century", 100 * 365 * 24 * 60 * 60 * 1000 * 1000 * 1000, { plural: "Centuries" });
addConversion(TimeUnit.Millennia, "Millennium", 1000 * 365 * 24 * 60 * 60 * 1000 * 1000 * 1000, { plural: "Millennia" });
/**
 * @param {?} unit
 * @param {?} singular
 * @param {?} nanos
 * @param {?=} __3
 * @return {?}
 */
function addConversion(unit, singular, nanos, { aliases = (/** @type {?} */ ([])), plural = singular + "s" } = {}) {
    for (const key of [
        unit,
        singular,
        singular.toLowerCase(),
        plural,
        plural.toLowerCase(),
        ...aliases,
    ]) {
        nanoConversions[key] = nanos;
    }
}

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
/**
 * Converts a 2D array to a csv string. Values are converted using micro-dash's toString().
 *
 * ```ts
 * toCsv([["a", "b", "c"], ["d", "e", "f"], ["g", "h", "i"]]); // "a,b,c\nd,e,f\ng,h,i"
 * toCsv([
 *   ["a", "", "string"]
 *   [undefined, null],
 *   [true, false],
 *   [1, 2, 3],
 *   [{}, { hi: "there" }]
 * ]) // "a,,string\n,\ntrue,false\n1,2,3\n[object Object],[object Object]"
 * ```
 * @param {?} content
 * @return {?}
 */
function toCsv(content) {
    return content
        .map((/**
     * @param {?} row
     * @return {?}
     */
    (row) => row.map((/**
     * @param {?} cell
     * @return {?}
     */
    (cell) => toCellString(cell))).join(",")))
        .join("\n");
}
/** @type {?} */
const specialCsvCharactersRegexp = /["|,|\n|\r]/;
/** @type {?} */
const allDoubleQuotes = /"/g;
/**
 * @param {?} value
 * @return {?}
 */
function toCellString(value) {
    /** @type {?} */
    const str = toString(value);
    if (specialCsvCharactersRegexp.test(str)) {
        return `"${str.replace(allDoubleQuotes, `""`)}"`;
    }
    else {
        return str;
    }
}

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
/**
 * Returns a new function to use in place of `func` that will call the provided hooks in addition to `func`. They are called in the following order:
 *
 * 1. `before`
 * 2. `original`
 * 3. `transform`
 * 4. `after`
 *
 * ```ts
 * const sum = (a: number, b: number) => a + b;
 * const sumAndLog = wrapFunction(sum, {
 *   after: (result, a, b) => {
 *     console.log(a, '+', b, '=', result);
 *   },
 * }
 * const sumPlusOne = wrapFunction(sum, {
 *   transform: (result) => result + 1,
 * });
 * ```
 * @template A, R, T
 * @param {?} original
 * @param {?} __1
 * @return {?}
 */
function wrapFunction(original, { before, transform, after, }) {
    /** @type {?} */
    const wrapped = (/**
     * @this {?}
     * @param {...?} args
     * @return {?}
     */
    function (...args) {
        if (before) {
            before.apply(this, args);
        }
        /** @type {?} */
        let result = original.apply(this, args);
        if (transform) {
            result = ((/** @type {?} */ (transform))).call(this, result, ...args);
        }
        if (after) {
            ((/** @type {?} */ (after))).call(this, result, ...args);
        }
        return result;
    });
    Object.defineProperty(wrapped, "length", { value: original.length });
    return wrapped;
}

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */

export { CallableObject, Deferred, TimeUnit, convertTime, createBuilder, elapsedToString, isDefined, isEqualAtDepth, roundToMultipleOf, sleep, toCsv, wrapFunction };
//# sourceMappingURL=s-js-utils.js.map
